__all__ = [
    "add",
    "diffuse",
    "init",
    "link",
    "list_",
    "remove",
    "rename",
    "unlink"
]


from .commands.add import add
from .commands.diffuse import diffuse
from .commands.init import init
from .commands.link import link
from .commands.list_ import list_
from .commands.remove import remove
from .commands.rename import rename
from .commands.unlink import unlink



from syndot.utils.colors import Color


def print_action(
    action_type: str,
    system_target_path: str,
    settings_target_path: str
) -> None:
    if action_type == 'link':
        action = 'Linking'
        preposition = 'to'
        print(
            f"{action} {Color.link(system_target_path)} "
            f"{preposition} {Color.settings(settings_target_path)}"
        )
    elif action_type == 'unlink':
        action = 'Unlinking'
        preposition = 'from'
        print(
            f"{action} {Color.link(system_target_path)} {preposition} "
            f"{Color.settings(settings_target_path)}"
        )
    elif action_type == 'diffuse':
        action = 'Diffusing'
        preposition = 'to'
        print(
            f"{action} {Color.settings(settings_target_path)} {preposition} "
            f"{Color.link(system_target_path)}"
        )


def print_relationship(
    system_target_path: str,
    settings_target_path: str,
    symbol: str
) -> None:
    print(
        f"{Color.link(system_target_path)} {Color.symbol(symbol)} "
        f"{Color.settings(settings_target_path)}"
    )


def print_highlight(sentence: str, end: str = '\n') -> None:
    print(f"{Color.highlight(sentence)}", end=end)


def print_error(error: str, end: str = '\n') -> None:
    print(f"{Color.error(error)}", end=end)



from configparser import ConfigParser
import os
from syndot.init_config import CONFIG_DIR_PATH
from syndot.utils.colors import Color


STYLE_PATH = os.path.join(
    CONFIG_DIR_PATH,
    '_templates',
    'style.ini'
)
if not os.path.exists(CONFIG_DIR_PATH) or not os.path.exists(STYLE_PATH):
    STYLE_PATH = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        '_templates',
        'style.ini'
    )

style_config = ConfigParser()
style_config.read(STYLE_PATH)

indicator = style_config['gum']['indicator'].replace('\'', '')
indicator_foreground = Color.INDICATOR_FOREGROUND
selected_prefix = style_config['gum']['selected_prefix'].replace('\'', '')
selected_indicator_foreground = Color.SELECTED_INDICATOR_FOREGROUND
unselected_prefix = style_config['gum']['unselected_prefix'].replace('\'', '')
match_foreground = Color.MATCH_FOREGROUND

GUM_FILTER_COMMAND = [
    'gum',
    'filter'
]
GUM_OPTIONS = [
    '--prompt= ',
    '--placeholder=Filter labels',
    '--width=100',
    '--no-limit',
    f"--indicator={indicator}",
    f"--indicator.foreground={indicator_foreground}",
    f"--selected-prefix={selected_prefix}",
    f"--selected-indicator.foreground={selected_indicator_foreground}",
    f"--unselected-prefix={unselected_prefix}",
    f"--match.foreground={match_foreground}"
]

if style_config['gum']['match_bold'] in ['True', 'true']:
    GUM_OPTIONS.append('--match.bold')


def complete_gum_filter_command(labels: list[str]) -> list[str]:
    GUM_FILTER_COMMAND.extend(labels)
    GUM_FILTER_COMMAND.extend(GUM_OPTIONS)

    return GUM_FILTER_COMMAND



from syndot.parser.parser import command_parser, CommandFormatter


init_parser = command_parser.add_parser(
    'init',
    prog="syndot init",
    usage="%(prog)s [[-p | --path] <PATH>]",
    description="Initialize settings directory where store dotfiles",
    help="Initialize settings directory where store dotfiles",
    add_help=False,
    formatter_class=CommandFormatter
)

init_parser.add_argument(
    '-p', '--path',
    required=False,
    dest='path',
    metavar='<PATH>',
    help="Path to the settings directory. If not provided, it creates the "
         "settings directory at the default path: ~/Settings"
)



from argparse import Namespace
from configparser import ConfigParser
import os
import subprocess
from syndot.utils.path import expand_home_path
from syndot.utils.system_info import gum_is_available
from syndot.utils.gum_style import complete_gum_filter_command


def read_map_file(map_file_path: str | None) -> ConfigParser:
    map_file_path = expand_home_path(
        map_file_path if map_file_path is not None else 'map.ini'
    )
    if not os.path.exists(map_file_path):
        if map_file_path == 'map.ini':
            raise FileNotFoundError(
                "Missing map.ini file in current directory."
            )
        else:
            raise FileNotFoundError(
                "Missing map.ini file at the specified path."
            )
    config = ConfigParser()
    config.read(map_file_path)

    return config


def get_map_info(
    config: ConfigParser,
    args: Namespace
) -> tuple[str, list[str]]:
    settings_dir = config['Path']['settings_dir']
    targets, unavailable_labels, unavailable_paths = _get_available_targets(
        config=config,
        args=args
    )

    _compose_error_message(
        unavailable_labels=unavailable_labels,
        unavailable_paths=unavailable_paths
    )

    if args.interactive:
        if gum_is_available():
            gum_filter_command = complete_gum_filter_command(
                labels=list(config['Targets'].keys())
            )
            selected_labels = subprocess.run(
                gum_filter_command,
                stdout=subprocess.PIPE
            ).stdout.decode('utf-8').split()
            targets = [path for selected_label in selected_labels
                       for path in config['Targets'][selected_label].split()]
        else:
            raise OSError("Interactive option requires 'gum', which is not "
                          "available in the system")
    else:
        if args.label is None and args.path is None:
            for target in config['Targets'].values():
                targets.extend(target.split())

    if args.start:
        targets = [target for target in targets
                   if target.startswith(args.start)]

    return settings_dir, targets


def write_map_file(map_file_path: str | None, config: ConfigParser) -> None:
    with open(map_file_path, 'w') as map_file:
        config.write(map_file)


def _get_available_targets(
    config: ConfigParser,
    args: Namespace
) -> tuple[list[str], [list[str], list[str]]]:
    targets = []

    available_labels = list(config['Targets'].keys())
    unavailable_labels = []
    if args.label is not None:
        for label in args.label:
            if label in available_labels:
                targets.extend(config['Targets'][label].split())
            else:
                unavailable_labels.append(label)

    available_paths = []
    for path in config['Targets'].values():
        available_paths.extend(path.split())
    available_paths = [expand_home_path(path) for path in available_paths]
    unavailable_paths = []
    if args.path is not None:
        for path in args.path:
            if path.endswith(os.sep):
                path = path[:-1]
            path = expand_home_path(path=path)
            if path in available_paths:
                targets.append(path)
            else:
                unavailable_paths.append(path)

    return targets, unavailable_labels, unavailable_paths


def _compose_error_message(
    unavailable_labels: list[str],
    unavailable_paths: list[str]
) -> None:
    error_message = ""
    if unavailable_labels:
        error_message += "\nThe following labels are not available in the map"\
                         "file:\n"
        for label in unavailable_labels:
            error_message += f"    {label}\n"
    if unavailable_paths:
        error_message += "\nThe following paths are not available in the map"\
                         "file:\n"
        for path in unavailable_paths:
            error_message += f"    {path}\n"
    if error_message:
        raise NameError(error_message)



from syndot import commands
from syndot.parser.parser import parser
from syndot.parser.remove_parser import remove_parser
from syndot import init_config


init_config.init_config()
command_map = {
    'init': commands.init.init,
    'link': commands.link.link,
    'unlink': commands.unlink.unlink,
    'diffuse': commands.diffuse.diffuse,
    'add': commands.add.add,
    'remove': commands.remove.remove,
    'rename': commands.rename.rename,
    'list': commands.list_.list_
}


def main():
    args = parser.parse_args()

    if args.command == 'remove' and args.label is None and args.path is None:
        remove_parser.error(
            "At least a [-l | --label] or a [-p | --path] must be specified"
        )

    command_map[args.command](args=args)


if __name__ == "__main__":
    main()



from syndot.parser.parser import command_parser, CommandFormatter


add_parser = command_parser.add_parser(
    'add',
    prog="syndot add",
    usage="%(prog)s ([-l | --label] <LABEL>) ([-p | --path] <PATH>...) "
          "[[-m | --mapfile] <MAP_FILE>]",
    description="Add dotfiles to the map file",
    help="Add dotfiles to the map file",
    add_help=False,
    formatter_class=CommandFormatter
)

add_parser.add_argument(
    '-l', '--label',
    required=True,
    dest='label',
    metavar='<LABEL>',
    help="Label under which to group multiple target paths relating to "
         "the same set of configurations. A single label must be provided"
)

add_parser.add_argument(
    '-p', '--path',
    required=True,
    nargs='+',
    dest='path',
    metavar='<PATH>',
    help="Dotfile path(s) to be added to the map file and grouped under "
         "<LABEL>. At least a path must be provided"
)



from syndot.parser.parser import command_parser, CommandFormatter


rename_parser = command_parser.add_parser(
    'rename',
    prog="syndot rename",
    usage="%(prog)s ([-o | --old] <OLD_LABEL>) ([-n | --new] <NEW_LABEL>) "
          "[[-m | --mapfile] <MAP_FILE>]",
    description="Rename existing label in the map file",
    help="Rename existing label in the map file",
    add_help=False,
    formatter_class=CommandFormatter
)

rename_parser.add_argument(
    '-n', '--new',
    required=True,
    dest='new_label',
    metavar='<NEW_LABEL>',
    help="New label to use"
)

rename_parser.add_argument(
    '-o', '--old',
    required=True,
    dest='old_label',
    metavar='<OLD_LABEL>',
    help="Existing label to rename"
)



from argparse import Namespace
import os
from syndot.utils.map_file import read_map_file, write_map_file
from syndot.utils.path import expand_home_path


def remove(args: Namespace) -> None:
    map_file_path = expand_home_path(
        args.mapfile if args.mapfile is not None else 'map.ini'
    )

    config = read_map_file(map_file_path=map_file_path)
    current_targets = dict(config['Targets'])
    for label, paths in current_targets.items():
        current_targets[label] = paths.split()

    if args.label is not None:
        for label in args.label:
            if label in current_targets.keys():
                current_targets.pop(label)
            else:
                raise NameError(f"Label {label} not found in map file")

    if args.path is not None:
        for path in args.path:
            path_removed = False
            if path.endswith(os.sep):
                path = path[:-1]
            for current_paths in current_targets.values():
                if path in current_paths:
                    current_paths.remove(path)
                    path_removed = True
            if not path_removed:
                raise NameError(f"Path {path} not found in map file")

    current_targets = dict(sorted(current_targets.items()))
    config['Targets'].clear()
    for label, paths in current_targets.items():
        paths = list(set(paths))
        paths.sort()
        config['Targets'][label] = '\n' + '\n'.join(paths)

    write_map_file(map_file_path=map_file_path, config=config)



import subprocess
from syndot.utils.colors import Color
from syndot.utils.system_info import gum_is_available


MESSAGE = 'Do you want to proceed?'
VALID_PROMPT_CHOICES = {
    'y': True,
    'ye': True,
    'yes': True,
    'n': False,
    'no': False
}
GUM_CONFIRM_COMMAND = [
    'gum',
    'confirm',
    f"{MESSAGE}",
    '--default=false',
    '--affirmative=YES',
    '--negative=NO',
    f"--prompt.foreground={Color.PROMPT_SENTENCE}",
    f"--selected.foreground={Color.PROMPT_BACKGROUND}",
    f"--selected.background={Color.PROMPT_FOREGROUND}",
    f"--unselected.foreground={Color.PROMPT_FOREGROUND}",
    f"--unselected.background={Color.PROMPT_BACKGROUND}"
]


def ask_to_proceed() -> bool:
    if gum_is_available():
        return __ask_to_proceed_gum()
    else:
        return __ask_to_proceed_prompt()


def __ask_to_proceed_prompt() -> bool:
    choice = ''
    while choice not in VALID_PROMPT_CHOICES:
        choice = input(f"{MESSAGE} (y/N)? ").lower()
        if choice == '':
            choice = 'n'

    return VALID_PROMPT_CHOICES[choice]


def __ask_to_proceed_gum() -> bool:
    choice = subprocess.run(
        '_'.join(GUM_CONFIRM_COMMAND).split('_'),
        stdout=subprocess.PIPE,
        text=True
    ).returncode

    return choice == 0



import os
from syndot.utils.path import expand_home_path
from syndot.utils.file_actions import copy


SOURCE_PATH = os.path.dirname(__file__)
HOME_PATH = expand_home_path('~')
CONFIG_DIR_PATH = os.path.join(HOME_PATH, '.config', 'syndot')
LOG_FILE_PATH = os.path.join(
    HOME_PATH,
    '.local',
    'share',
    'syndot',
    'log_file.log'
)


def copy_sources(source: str, destination: str) -> None:
    source_path = os.path.join(SOURCE_PATH, source)
    destination_path = os.path.join(CONFIG_DIR_PATH, destination)
    copy(source_path, destination_path)


def init_config() -> None:
    if not os.path.exists(CONFIG_DIR_PATH):
        copy_sources(source='_templates', destination='templates')
        copy_sources(source='_colorschemes', destination='colorschemes')

    if not os.path.exists(LOG_FILE_PATH):
        os.makedirs(os.path.dirname(LOG_FILE_PATH))
        with open(LOG_FILE_PATH, 'w') as log_file:
            log_file.write('')



from syndot.parser.parser import command_parser, CommandFormatter


list_parser = command_parser.add_parser(
    'list',
    prog="syndot list",
    usage="%(prog)s [-d | --directory] [[-l | --label] | [-p | --path]] | "
          "[[-s | --search] <SEARCH>] [[-m | --mapfile] <MAP_FILE>]",
    description="List dotfiles in the map file",
    help="List dotfiles in the map file",
    add_help=False,
    formatter_class=CommandFormatter
)

list_targets = list_parser.add_mutually_exclusive_group(required=False)

list_parser.add_argument(
    '-d', '--directory',
    action='store_true',
    default=False,
    required=False,
    dest='directory',
    help="Print the settings directory"
)

list_targets.add_argument(
    '-l', '--label',
    action='store_true',
    default=False,
    required=False,
    dest='label',
    help="List only target labels. Not allowed together with the "
         "[-p | --path] or [-s | --search] options"
)

list_targets.add_argument(
    '-p', '--path',
    action='store_true',
    default=False,
    required=False,
    dest='path',
    help="List only target paths. Now allowed together with the "
         "[-l | --label] or [-s | --search] options"
)

list_targets.add_argument(
    '-s', '--search',
    required=False,
    dest='search',
    metavar='<SEARCH>',
    help="List paths of the specified <SEARCH> label. Not allowed together "
         "with the [-l | --label] or [-p | --path] options"
)



from syndot.parser.parser import command_parser, CommandFormatter


remove_parser = command_parser.add_parser(
    'remove',
    prog="syndot remove",
    usage="%(prog)s ([[-l | --label] <LABEL>...] | [[-p | --path] <PATH>...]) "
          "[[-m | --mapfile] <MAP_FILE>]",
    description="Remove dotfiles from the map file",
    help="Remove dotfiles from the map file",
    add_help=False,
    formatter_class=CommandFormatter
)

remove_parser.add_argument(
    '-l', '--label',
    required=False,
    nargs='+',
    dest='label',
    metavar='<LABEL>',
    help="Label(s) and relative path(s) to remove from the map file. One or "
         "more labels can be provided. At least a <LABEL> or a <PATH> must be "
         "provided"
)

remove_parser.add_argument(
    '-p', '--path',
    required=False,
    nargs='+',
    dest='path',
    metavar='<PATH>',
    help="Dotfile path(s) to remove from the map file. One of more paths can "
         "be provided. At least a <LABEL> or a <PATH> must be provided"
)



import os
from syndot.utils.print_ import print_highlight, print_relationship


def skip_dotfiles(
    targets_list: list[str],
    many_targets_sentence: str,
    single_file_sentence: str,
    single_directory_sentence: str
) -> None:
    if targets_list:
        for target in targets_list:
            print(f"Skip {target}")
        if len(targets_list) > 1:
            print_highlight(many_targets_sentence)
        else:
            if os.path.isfile(targets_list[0]):
                print_highlight(single_file_sentence)
            else:
                print_highlight(single_directory_sentence)
        print()


def print_dotfiles_to_manage(
    targets_list: dict[str, str],
    many_targets_sentence: str,
    single_file_sentence: str,
    single_directory_sentence: str,
    symbol: str
) -> None:
    n_targets = len(targets_list.keys())
    if n_targets > 1:
        for system_target_path, settings_target_path in targets_list.items():
            print_relationship(
                system_target_path=system_target_path,
                settings_target_path=settings_target_path,
                symbol=symbol
            )
        print_highlight(many_targets_sentence)
    else:
        print_relationship(
            system_target_path=list(targets_list.keys())[0],
            settings_target_path=list(targets_list.values())[0],
            symbol=symbol
        )
        if os.path.isfile(list(targets_list.keys())[0]):
            print_highlight(single_file_sentence)
        else:
            print_highlight(single_directory_sentence)



import os


def generate_backup_path(path: str) -> str:
    backup_extension = ''
    if os.path.isfile(path):
        backup_extension = '.bak'
    elif os.path.isdir(path):
        backup_extension = '_bak'
    parent, target = os.path.split(path)
    return os.path.join(parent, target + backup_extension)


def split_path(path: str) -> list[str]:
    return path.split(os.path.sep)[1:]


def expand_home_path(path: str) -> str:
    sudo_user = os.getenv('SUDO_USER')
    if sudo_user:
        if '~' in path:
            home = os.path.join(os.path.sep, 'home', sudo_user)
            return path.replace('~', home)
        else:
            return path
    else:
        return os.path.expanduser(path)


def compose_target_paths(settings_dir: str, target: str) -> tuple[str, str]:
    system_target_path = expand_home_path(target)
    settings_dir = expand_home_path(settings_dir)
    settings_target_path = os.path.join(
        settings_dir, *split_path(system_target_path))
    return system_target_path, settings_target_path



from argparse import Namespace
from syndot.utils.map_file import read_map_file
from syndot.utils.path import expand_home_path
from syndot.utils.print_ import print_highlight
from syndot.utils.colors import Color


def list_(args: Namespace) -> None:
    config = read_map_file(map_file_path=args.mapfile)

    settings_dir = expand_home_path(config['Path']['settings_dir'])
    targets = dict(config['Targets'])

    if args.label:
        for label in targets.keys():
            _list_labels(label=label)
        if args.directory:
            _print_settings_directory(settings_dir=settings_dir)
        return

    if args.path:
        for paths in targets.values():
            _list_paths(paths=paths, indent=False)
        if args.directory:
            _print_settings_directory(settings_dir=settings_dir)
        return

    if args.search:
        if args.search not in targets.keys():
            raise NameError(f"Label {args.search} not found in map file")

        _list_labels(label=args.search)
        _list_paths(paths=targets[args.search], indent=True)
        if args.directory:
            _print_settings_directory(settings_dir=settings_dir)
        return

    for label, paths in targets.items():
        _list_labels(label=label)
        _list_paths(paths=paths, indent=True)
        print()
    if args.directory:
        _print_settings_directory(settings_dir=settings_dir)


def _list_labels(label: str) -> None:
    print(label.split()[0])


def _list_paths(paths: str, indent: bool) -> None:
    paths = paths.split()
    for path in paths:
        if indent:
            print(f"    {path}")
        else:
            print(path)


def _print_settings_directory(settings_dir: str) -> None:
    print_highlight("Settings directory:", end=' ')
    print(Color.settings(settings_dir))



from argparse import Namespace
import os
from syndot.utils.map_file import read_map_file, write_map_file
from syndot.utils.path import expand_home_path


def add(args: Namespace) -> None:
    map_file_path = expand_home_path(
        args.mapfile if args.mapfile is not None else 'map.ini'
    )

    config = read_map_file(map_file_path=map_file_path)
    current_targets = dict(config['Targets'])
    for label, paths in current_targets.items():
        current_targets[label] = paths.split()

    for target_path in args.path:
        if target_path.endswith(os.sep):
            target_path = target_path[:-1]

        if not os.path.exists(target_path):
            raise OSError(f"Target {target_path} not found")

        target_path = expand_home_path(target_path)

        if _check_path_already_under_label(
            label=args.label,
            target_path=target_path,
            current_targets=current_targets
        ):
            continue

        if _check_path_already_under_other_label(
            target_path=target_path,
            current_targets=current_targets
        ):
            continue

        if args.label not in current_targets.keys():
            current_targets[args.label] = []
        current_targets[args.label].append(target_path)

    if args.label in current_targets.keys():
        current_targets[args.label] = list(set(current_targets[args.label]))
        current_targets[args.label].sort()
    current_targets = dict(sorted(current_targets.items()))
    config['Targets'].clear()
    for label, paths in current_targets.items():
        config['Targets'][label] = '\n' + '\n'.join(paths)

    write_map_file(map_file_path=map_file_path, config=config)


def _check_path_already_under_label(
    label: str,
    target_path: str,
    current_targets: dict[str, list[str]]
) -> bool:
    if (label in current_targets.keys()) and \
       (target_path in current_targets[label]):
        print(f"Target {target_path} already in map file associated to label "
              f"{label!r}")
        return True
    else:
        return False


def _check_path_already_under_other_label(
    target_path: str,
    current_targets: dict[str, list[str]]
) -> bool:
    if target_path in [path for target in current_targets.values()
                       for path in target]:
        label = _find_label_associated_to_path(
            target_path=target_path,
            current_targets=current_targets
        )
        print(f"Target {target_path} already in map file associated to label "
              f"{label!r}")
        return True
    else:
        return False


def _find_label_associated_to_path(
    target_path: str,
    current_targets: dict[str, list[str]]
) -> str:
    for label, paths in current_targets.items():
        if target_path in paths:
            return label
    return ' '



from syndot.parser.parser import command_parser, CommandFormatter


link_parser = command_parser.add_parser(
    'link',
    prog="syndot link",
    usage="%(prog)s [-b | --backup] ([-i | --interactive] | "
          "[[-l | --label] <LABEL>...] | [[-p | --path] <PATH>...]) "
          "[[-m | --mapfile] <MAP_FILE>] [-n | --no-confirm] "
          "[[-s | --start] <PATH_START>] ",
    description="Move dotfiles to settings directory "
                "and create symlinks to them",
    help="Move dotfiles to settings directory and create symlinks to them",
    add_help=False,
    formatter_class=CommandFormatter
)

link_parser.add_argument(
    '-b', '--backup',
    action='store_true',
    default=False,
    required=False,
    dest='backup',
    help="Create a backup copy of the original dotfile"
)

targets_group = link_parser.add_mutually_exclusive_group(required=False)

targets_group.add_argument(
    '-i', '--interactive',
    action='store_true',
    default=False,
    required=False,
    dest='interactive',
    help="Select label(s) to link the associated path(s) in interactive mode "
         "using gum, if it is installed in the system. At least a label must "
         "be selected. Not allowed together with [-l | --label] or "
         "[-p | --path] options"
)

targets_group.add_argument(
    '-l', '--label',
    required=False,
    nargs='+',
    dest='label',
    metavar="<LABEL>",
    help="Label(s) to link the associated path(s). At least a <LABEL> must be "
         "provided. Not allowed together with [-i | --interactive] or "
         "[-p | --path] options"
)

targets_group.add_argument(
    '-p', '--path',
    required=False,
    nargs='+',
    dest='path',
    metavar='<PATH>',
    help="Dotfile path(s) to link. At least a <PATH> must be provided. Not "
         "allowed together with [-i | --interactive] or [-l | --label] options"
)



import os
from pytest import mark
from syndot import init_config
from tests.conftest import TEST_DATA_PATH, reset_environment


@mark.config
class TestInitConfig:

    @mark.genuine
    def test_function(self, monkeypatch):
        reset_environment()

        def mock_config_path():
            return os.path.join(TEST_DATA_PATH, '.config', 'syndot')

        def mock_log_path():
            return os.path.join(
                TEST_DATA_PATH, '.local', 'share', 'syndot', 'log_file.log')

        config_path = mock_config_path()
        log_file_path = mock_log_path()

        monkeypatch.setattr(init_config, 'CONFIG_DIR_PATH', config_path)
        monkeypatch.setattr(init_config, 'LOG_FILE_PATH', log_file_path)

        assert not os.path.exists(config_path)

        init_config.init_config()

        assert os.path.exists(config_path)
        assert os.path.isdir(config_path)
        for asset in ['templates', 'colorschemes']:
            asset_path = os.path.join(config_path, asset)
            assert os.path.exists(asset_path)
            assert os.path.isdir(asset_path)
            assert os.listdir(asset_path)
        assert os.path.exists(log_file_path)
        assert os.path.isfile(log_file_path)

        reset_environment()



from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import (
    one_of,
    none,
    booleans,
    text,
    characters,
    sampled_from,
    lists
)
from pytest import mark, raises
from syndot.parser.parser import parser
from tests.conftest import paths


@mark.parsers
class TestParser:

    @mark.genuine
    @given(path=one_of(paths(), none()))
    @settings(max_examples=100, deadline=None)
    def test_init(self, path):
        input_arguments = ['init']
        if path:
            input_arguments.extend(['-p', path])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'init'
        if path:
            assert parsed_arguments.path == path
        else:
            assert parsed_arguments.path is None

    @mark.genuine
    @given(
        backup=booleans(),
        label_list=one_of(
            lists(
                min_size=1,
                max_size=5,
                elements=text(
                    min_size=5,
                    max_size=10,
                    alphabet=characters(min_codepoint=97, max_codepoint=122)
                )
            ),
            none()
        ),
        path_list=one_of(
            lists(
                min_size=1,
                max_size=5,
                elements=text(
                    min_size=5,
                    max_size=10,
                    alphabet=characters(min_codepoint=97, max_codepoint=122)
                )
            ),
            none()
        ),
        map_file_path=one_of(paths(), none()),
        no_confirm=booleans(),
        interactive=booleans(),
        start=one_of(
            text(
                min_size=5,
                max_size=10,
                alphabet=characters(min_codepoint=97, max_codepoint=122)
            ),
            none()
        )
    )
    @settings(max_examples=100, deadline=None)
    def test_link(
        self,
        backup,
        label_list,
        path_list,
        map_file_path,
        no_confirm,
        interactive,
        start
    ):
        input_arguments = ['link']
        if backup:
            input_arguments.append('-b')
        if label_list:
            if path_list:
                if interactive:
                    input_arguments.extend(['-i'])
            else:
                input_arguments.extend(['-l', *label_list])
        else:
            if path_list:
                input_arguments.extend(['-p', *path_list])
            else:
                if interactive:
                    input_arguments.extend(['-i'])
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if no_confirm:
            input_arguments.extend(['-n'])
        if start:
            input_arguments.extend(['-s', start])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'link'
        assert parsed_arguments.backup == backup
        if label_list:
            if path_list:
                assert parsed_arguments.label is None
                assert parsed_arguments.path is None
                assert parsed_arguments.interactive == interactive
            else:
                assert parsed_arguments.label == label_list
                assert parsed_arguments.path is None
                assert not parsed_arguments.interactive
        else:
            if path_list:
                assert parsed_arguments.label is None
                assert parsed_arguments.path == path_list
                assert not parsed_arguments.interactive
            else:
                assert parsed_arguments.label is None
                assert parsed_arguments.path is None
                assert parsed_arguments.interactive == interactive
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        assert parsed_arguments.no_confirm == no_confirm
        if start:
            assert parsed_arguments.start == start
        else:
            assert parsed_arguments.start is None

    @mark.genuine
    @given(
        label_list=one_of(
            lists(
               min_size=1,
               max_size=5,
               elements=text(
                   min_size=5,
                   max_size=10,
                   alphabet=characters(min_codepoint=97, max_codepoint=122)
                )
            ),
            none()
        ),
        path_list=one_of(
            lists(
                min_size=1,
                max_size=5,
                elements=text(
                    min_size=5,
                    max_size=10,
                    alphabet=characters(min_codepoint=97, max_codepoint=122)
                )
            ),
            none()
        ),
        map_file_path=one_of(paths(), none()),
        no_confirm=booleans(),
        interactive=booleans(),
        start=one_of(
            text(
                min_size=5,
                max_size=10,
                alphabet=characters(min_codepoint=97, max_codepoint=122)
            ),
            none()
        )
    )
    @settings(max_examples=100, deadline=None)
    def test_unlink(
        self,
        label_list,
        path_list,
        map_file_path,
        no_confirm,
        interactive,
        start
    ):
        input_arguments = ['unlink']
        if label_list:
            if path_list:
                if interactive:
                    input_arguments.extend(['-i'])
            else:
                input_arguments.extend(['-l', *label_list])
        else:
            if path_list:
                input_arguments.extend(['-p', *path_list])
            else:
                if interactive:
                    input_arguments.extend(['-i'])
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if no_confirm:
            input_arguments.extend(['-n'])
        if start:
            input_arguments.extend(['-s', start])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'unlink'
        if label_list:
            if path_list:
                assert parsed_arguments.label is None
                assert parsed_arguments.path is None
                assert parsed_arguments.interactive == interactive
            else:
                assert parsed_arguments.label == label_list
                assert parsed_arguments.path is None
                assert not parsed_arguments.interactive
        else:
            if path_list:
                assert parsed_arguments.label is None
                assert parsed_arguments.path == path_list
                assert not parsed_arguments.interactive
            else:
                assert parsed_arguments.label is None
                assert parsed_arguments.path is None
                assert parsed_arguments.interactive == interactive
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        assert parsed_arguments.no_confirm == no_confirm
        if start:
            assert parsed_arguments.start == start
        else:
            assert parsed_arguments.start is None

    @mark.genuine
    @given(
        label_list=one_of(
            lists(
               min_size=1,
               max_size=5,
               elements=text(
                   min_size=5,
                   max_size=10,
                   alphabet=characters(min_codepoint=97, max_codepoint=122)
                )
            ),
            none()
        ),
        path_list=one_of(
            lists(
                min_size=1,
                max_size=5,
                elements=text(
                    min_size=5,
                    max_size=10,
                    alphabet=characters(min_codepoint=97, max_codepoint=122)
                )
            ),
            none()
        ),
        map_file_path=one_of(paths(), none()),
        no_confirm=booleans(),
        interactive=booleans(),
        start=one_of(
            text(
                min_size=5,
                max_size=10,
                alphabet=characters(min_codepoint=97, max_codepoint=122)
            ),
            none()
        )
    )
    @settings(max_examples=100, deadline=None)
    def test_diffuse(
        self,
        label_list,
        path_list,
        map_file_path,
        no_confirm,
        interactive,
        start
    ):
        input_arguments = ['diffuse']
        if label_list:
            if path_list:
                if interactive:
                    input_arguments.extend(['-i'])
            else:
                input_arguments.extend(['-l', *label_list])
        else:
            if path_list:
                input_arguments.extend(['-p', *path_list])
            else:
                if interactive:
                    input_arguments.extend(['-i'])
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if no_confirm:
            input_arguments.extend(['-n'])
        if start:
            input_arguments.extend(['-s', start])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'diffuse'
        if label_list:
            if path_list:
                assert parsed_arguments.label is None
                assert parsed_arguments.path is None
                assert parsed_arguments.interactive == interactive
            else:
                assert parsed_arguments.label == label_list
                assert parsed_arguments.path is None
                assert not parsed_arguments.interactive
        else:
            if path_list:
                assert parsed_arguments.label is None
                assert parsed_arguments.path == path_list
                assert not parsed_arguments.interactive
            else:
                assert parsed_arguments.label is None
                assert parsed_arguments.path is None
                assert parsed_arguments.interactive == interactive
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        assert parsed_arguments.no_confirm == no_confirm
        if start:
            assert parsed_arguments.start == start
        else:
            assert parsed_arguments.start is None

    @given(
        label=text(
            min_size=5,
            max_size=10,
            alphabet=characters(min_codepoint=97, max_codepoint=122)
        ),
        map_file_path=one_of(paths(), none()),
        path_list=lists(
            min_size=5,
            max_size=10,
            elements=text(
                min_size=5,
                max_size=10,
                alphabet=characters(min_codepoint=97, max_codepoint=122)
            )
        )
    )
    @settings(max_examples=100, deadline=None)
    def test_add(self, label, map_file_path, path_list):
        input_arguments = ['add']
        input_arguments.extend(['-l', label])
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        input_arguments.extend(['-p', *path_list])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'add'
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        assert parsed_arguments.label == label
        assert parsed_arguments.path == path_list

    @mark.genuine
    @given(
        label_list=lists(
            min_size=5,
            max_size=10,
            elements=text(
                min_size=5,
                max_size=10,
                alphabet=characters(min_codepoint=97, max_codepoint=122)
            )
        ),
        map_file_path=one_of(paths(), none()),
        path_list=lists(
            min_size=5,
            max_size=10,
            elements=text(
                min_size=5,
                max_size=10,
                alphabet=characters(min_codepoint=97, max_codepoint=122)
            )
        ),
        arg_is_label=booleans()
    )
    @settings(max_examples=100, deadline=None)
    def test_remove(self, label_list, map_file_path, path_list, arg_is_label):
        input_arguments = ['remove']
        if arg_is_label:
            input_arguments.extend(['-l', *label_list])
        else:
            input_arguments.extend(['-p', *path_list])
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'remove'
        if arg_is_label:
            assert parsed_arguments.label == label_list
            assert parsed_arguments.path is None
        else:
            assert parsed_arguments.label is None
            assert parsed_arguments.path == path_list
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None

    @mark.genuine
    @given(
        directory=booleans(),
        label=booleans(),
        map_file_path=one_of(paths(), none()),
        path=booleans()
    )
    @settings(max_examples=100, deadline=None)
    def test_list(self, directory, label, map_file_path, path):
        input_arguments = ['list']
        if directory:
            input_arguments.append('-d')
        if label and not path:
            input_arguments.append('-l')
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if path and not label:
            input_arguments.append('-p')

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'list'
        if directory:
            assert parsed_arguments.directory
        else:
            assert not parsed_arguments.directory
        if label and not path:
            assert parsed_arguments.label
            assert not parsed_arguments.path
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        if path and not label:
            assert parsed_arguments.path
            assert not parsed_arguments.label
        if not label and not path:
            assert not parsed_arguments.label
            assert not parsed_arguments.path

    @mark.genuine
    @given(
        old_label=text(
            min_size=5,
            max_size=10,
            alphabet=characters(min_codepoint=97, max_codepoint=122)
        ),
        new_label=text(
            min_size=5,
            max_size=10,
            alphabet=characters(min_codepoint=97, max_codepoint=122)
        )
    )
    @settings(max_examples=100, deadline=None)
    def test_rename(self, old_label, new_label):
        input_arguments = f"rename -o {old_label} -n {new_label}".split()

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'rename'
        assert parsed_arguments.old_label == old_label
        assert parsed_arguments.new_label == new_label

    @mark.genuine
    @given(abbreviation=booleans())
    @settings(
        max_examples=100,
        deadline=None,
        suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_version(self, abbreviation, capsys):
        args = ['-v'] if abbreviation else ['--version']
        with raises(SystemExit):
            parser.parse_args(args=args)
        printed_output = capsys.readouterr().out

        assert printed_output.startswith('syndot')

    @mark.genuine
    @given(abbreviation=booleans())
    @settings(
        max_examples=100,
        deadline=None,
        suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_help(self, abbreviation, capsys):
        args = ['-h'] if abbreviation else ['--help']
        with raises(SystemExit):
            parser.parse_args(args=args)
        printed_output = capsys.readouterr().out

        assert printed_output.startswith(
            "usage: syndot <COMMAND> [<OPTIONS>...]"
        )
        assert printed_output.endswith("Config file path: ~/.config/syndot\n")
        assert 'options:' in printed_output
        assert 'commands:' in printed_output
        assert '-h, --help' in printed_output
        assert '-v, --version' in printed_output
        assert 'add' in printed_output
        assert 'diffuse' in printed_output
        assert 'init' in printed_output
        assert 'link' in printed_output
        assert 'list' in printed_output
        assert 'remove' in printed_output
        assert 'rename' in printed_output
        assert 'unlink' in printed_output

    @mark.genuine
    @given(
        command=sampled_from(
            elements=[
                'add',
                'diffuse',
                'init',
                'link',
                'list',
                'remove',
                'unlink'
            ]
        ),
        abbreviation=booleans()
    )
    @settings(
        max_examples=100,
        deadline=None,
        suppress_health_check=[HealthCheck.function_scoped_fixture]
    )
    def test_command_help(self, command, abbreviation, capsys):
        args = [command]
        if abbreviation:
            args.append('-h')
        else:
            args.append('--help')
        with raises(SystemExit):
            parser.parse_args(args=args)
        printed_output = capsys.readouterr().out

        assert printed_output.startswith(f"usage: syndot {command}")
        assert 'options' in printed_output
        assert '-h, --help' in printed_output
        if command in ['add', 'diffuse', 'link', 'list', 'remove', 'unlink']:
            assert '-m, --mapfile <MAP_FILE>' in printed_output
        if command in ['diffuse', 'link', 'unlink']:
            assert '-l, --label' in printed_output
            assert '-p, --path' in printed_output
        if command == 'init':
            assert '-p, --path <PATH>' in printed_output
        if command == 'link':
            assert '-b, --backup' in printed_output



from syndot.parser.parser import command_parser, CommandFormatter


diffuse_parser = command_parser.add_parser(
    'diffuse',
    prog="syndot diffuse",
    usage="%(prog)s ([-i | --interactive] | [[-l | --label] <LABEL>...] | "
          "[[-p | --path] <PATH>...]) [[-m | --mapfile] <MAP_FILE>] "
          "[-n | --no-confirm] [[-s | --start] <PATH_START>]",
    description="Create dotfiles symlinks",
    help="Create dotfiles symlinks",
    add_help=False,
    formatter_class=CommandFormatter
)

targets_group = diffuse_parser.add_mutually_exclusive_group(required=False)

targets_group.add_argument(
    '-i', '--interactive',
    action='store_true',
    default=False,
    required=False,
    dest='interactive',
    help="Select label(s) to diffuse the associated path(s) in interactive "
         "mode using gum, if it is installed in the system. At least a label "
         "must be selected. Not allowed together with [-l | --label] or "
         "[-p | --path] options"
)

targets_group.add_argument(
    '-l', '--label',
    required=False,
    nargs='+',
    dest='label',
    metavar='<LABEL>',
    help="Label(s) to diffuse the associated path(s). At least a <LABEL> must "
         "be provided. Not allowed together with [-i | --interactive] or "
         "[-p | --path] options"
)

targets_group.add_argument(
    '-p', '--path',
    required=False,
    nargs='+',
    dest='path',
    metavar='<PATH>',
    help="Dotfile path(s) to diffuse. At least a <PATH> must be provided. Not "
         "allowed together with [-i | --interactive] or [-l | --label] options"
)



from argparse import Namespace
import os
from syndot.init_config import LOG_FILE_PATH
from syndot.utils.commands import skip_dotfiles, print_dotfiles_to_manage
from syndot.utils.file_actions import (
    change_parent_owner,
    change_child_owner,
    copy,
    remove
)
from syndot.utils.logger import log_error
from syndot.utils.map_file import get_map_info, read_map_file
from syndot.utils.path import compose_target_paths, generate_backup_path
from syndot.utils.print_ import (
    print_action,
    print_error,
    print_highlight
)
from syndot.utils.prompt import ask_to_proceed


def link(args: Namespace) -> None:
    settings_dir, targets = get_map_info(
        config=read_map_file(map_file_path=args.mapfile),
        args=args
    )

    targets_to_be_linked = {}
    already_existing_settings = {}
    missing_system_targets = []
    already_linked_targets = []
    corrupted_targets = []
    wrong_existing_links = []

    print_highlight("Looking for files and directories to link...")

    for target in targets:
        system_target_path, settings_target_path = compose_target_paths(
            settings_dir=settings_dir,
            target=target
        )

        if _check_tartgets_to_be_linked(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            targets_to_be_linked[system_target_path] = settings_target_path

        if _check_already_existing_settings(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            already_existing_settings[system_target_path] = \
                settings_target_path

        if _check_missing_system_targets(
            system_target_path=system_target_path
        ):
            missing_system_targets.append(system_target_path)

        if _check_already_linked_targets(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            already_linked_targets.append(system_target_path)

        if _check_corrupted_targets(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            corrupted_targets.append(system_target_path)

        if _check_wrong_existing_links(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            wrong_existing_links.append(system_target_path)

    if not any([
        targets_to_be_linked,
        already_existing_settings,
        missing_system_targets,
        already_linked_targets,
        corrupted_targets,
        wrong_existing_links
    ]):
        print_highlight("No files or directories found to link.")

    link_dotfiles(
        targets_list=targets_to_be_linked,
        settings_dir=settings_dir,
        backup=args.backup,
        many_targets_sentence=f"Found {len(targets_to_be_linked.keys())} "
                              f"files or directories to link.\nThey will be "
                              f"moved to settings directory and will be "
                              f"replaced by symbolic links.",
        single_file_sentence=f"Found {len(targets_to_be_linked.keys())} file "
                             f"to link.\nIt will be moved to settings "
                             f"directory and will be replaced by a symbolic "
                             f"link.",
        single_directory_sentence=f"Found {len(targets_to_be_linked.keys())} "
                                  f"directory to link.\nIt will be moved to "
                                  f"settings directory and will be replaced "
                                  f"by a symbolic link.",
        remove_settings=False,
        ask_for_confirmation=not args.no_confirm
    )

    link_dotfiles(
        targets_list=already_existing_settings,
        settings_dir=settings_dir,
        backup=args.backup,
        many_targets_sentence=f"Found {len(already_existing_settings.keys())} "
                              f"already existing files or directories in "
                              f"settings directory.\nThey will be removed, "
                              f"then the respective files and directories "
                              f"from the system will be moved to settings "
                              f"directory and will be replaced by symbolic "
                              f"links.",
        single_file_sentence=f"Found {len(already_existing_settings.keys())} "
                             f"already existing file in settings directory.\n"
                             f"It will be removed, then the respective file "
                             f"from the system will be moved to settings "
                             f"directory and will be replaced by a symbolic "
                             f"link.",
        single_directory_sentence=f"Found "
                                  f"{len(already_existing_settings.keys())} "
                                  f"already existing directory in settings "
                                  f"directory.\nIt will be removed, then the "
                                  f"respective directory from the system will "
                                  f"be moved to settings directory and will "
                                  f"be replaced by a symbolic link.",
        remove_settings=True,
        ask_for_confirmation=not args.no_confirm
    )

    skip_dotfiles(
        targets_list=missing_system_targets,
        many_targets_sentence=f"Skipping {len(missing_system_targets)} "
                              f"missing files or directories.",
        single_file_sentence=f"Skipping {len(missing_system_targets)} "
                             f"missing file.",
        single_directory_sentence=f"Skipping {len(missing_system_targets)} "
                                  f"missing directory."
    )

    skip_dotfiles(
        targets_list=already_linked_targets,
        many_targets_sentence=f"Skipping {len(already_linked_targets)} "
                              f"already linked files or directories.",
        single_file_sentence=f"Skipping {len(already_linked_targets)} "
                             f"already linked file.",
        single_directory_sentence=f"Skipping {len(already_linked_targets)} "
                                  f"already linked directory."
    )

    skip_dotfiles(
        targets_list=corrupted_targets,
        many_targets_sentence=f"Skipping {len(corrupted_targets)} files or "
                              f"directories because are links to non-existing "
                              f"files or directories.",
        single_file_sentence=f"Skipping {len(corrupted_targets)} file because "
                             f"is a link to a non-existing file.",
        single_directory_sentence=f"Skipping {len(corrupted_targets)} "
                                  f"directory because is a link to a "
                                  f"non-existing directory."
    )

    skip_dotfiles(
        targets_list=wrong_existing_links,
        many_targets_sentence=f"Skipping {len(wrong_existing_links)} files "
                              f"or directories because are links to wrong "
                              f"existing files or directories.",
        single_file_sentence=f"Skipping {len(wrong_existing_links)} file "
                             f"because is a link to a wrong existing file.",
        single_directory_sentence=f"Skipping {len(wrong_existing_links)} "
                                  f"directory because is a link to a wrong "
                                  f"existing directory."
    )


def link_dotfiles(
    targets_list: dict[str, str],
    settings_dir: str,
    backup: bool,
    many_targets_sentence: str,
    single_file_sentence: str,
    single_directory_sentence: str,
    remove_settings: bool,
    ask_for_confirmation: bool
) -> None:
    if targets_list:
        n_targets = len(targets_list.keys())

        print_dotfiles_to_manage(
            targets_list=targets_list,
            many_targets_sentence=many_targets_sentence,
            single_file_sentence=single_directory_sentence,
            single_directory_sentence=single_directory_sentence,
            symbol='-->'
        )

        proceed = ask_to_proceed() if ask_for_confirmation else True

        if proceed:
            for i, (system_target_path, settings_target_path) in \
                    enumerate(targets_list.items(), 1):
                try:
                    print_action(
                        action_type='link',
                        system_target_path=system_target_path,
                        settings_target_path=settings_target_path
                    )
                    print(f"Total ({i}/{n_targets})", end='\r')

                    if remove_settings:
                        remove(path=settings_target_path)
                    copy(
                        source=system_target_path,
                        destination=settings_target_path
                    )
                    change_parent_owner(
                        source=system_target_path,
                        destination=settings_target_path,
                        settings_dir=settings_dir
                    )
                    if os.path.isdir(system_target_path):
                        change_child_owner(
                            source=system_target_path,
                            destination=settings_target_path
                        )
                    if backup:
                        backup_path = generate_backup_path(
                                path=system_target_path
                        )
                        os.rename(system_target_path, backup_path)
                    else:
                        remove(system_target_path)
                    os.symlink(settings_target_path, system_target_path)
                except Exception as e:
                    print_error(
                        f"Error in linking {system_target_path}. "
                        f"Check {LOG_FILE_PATH} for more details."
                    )
                    log_error(f"{e}")
        print('\n')


def _check_tartgets_to_be_linked(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return not os.path.islink(system_target_path) and \
           os.path.exists(system_target_path) and \
           not os.path.exists(settings_target_path)


def _check_already_existing_settings(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return not os.path.islink(system_target_path) and \
           os.path.exists(system_target_path) and \
           os.path.exists(settings_target_path)


def _check_missing_system_targets(system_target_path: str) -> bool:
    return not os.path.islink(system_target_path) and \
           not os.path.exists(system_target_path)


def _check_already_linked_targets(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return os.path.islink(system_target_path) and \
           (os.readlink(system_target_path) == settings_target_path) and \
           os.path.exists(system_target_path)


def _check_corrupted_targets(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return os.path.islink(system_target_path) and \
           (os.readlink(system_target_path) == settings_target_path) and \
           not os.path.exists(system_target_path)


def _check_wrong_existing_links(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return os.path.islink(system_target_path) and \
           (os.readlink(system_target_path) != settings_target_path)



from argparse import Namespace
from configparser import ConfigParser
import os
from syndot.init_config import CONFIG_DIR_PATH
from syndot.utils.path import expand_home_path
from syndot.utils.map_file import read_map_file, write_map_file


DEFAULT_SETTINGS_DIR = '~/Settings'
MAP_TEMPLATE_PATH = os.path.join(CONFIG_DIR_PATH, 'templates', 'map.ini')
if (not os.path.exists(CONFIG_DIR_PATH) or
        not os.path.exists(MAP_TEMPLATE_PATH)):
    MAP_TEMPLATE_PATH = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        '_templates',
        'map.ini'
    )


def init(args: Namespace) -> None:
    settings_dir = expand_home_path(
        args.path if args.path is not None else DEFAULT_SETTINGS_DIR
    )
    if os.path.exists(settings_dir):
        raise ValueError(f"Settings directory {settings_dir} already exists.")
    os.makedirs(settings_dir)

    config = read_map_file(MAP_TEMPLATE_PATH)
    config['Path']['settings_dir'] = settings_dir

    _expand_default_target_path(config=config)

    write_map_file(
        map_file_path=os.path.join(settings_dir, 'map.ini'),
        config=config
    )


def _expand_default_target_path(config: ConfigParser) -> None:
    for label, paths in config['Targets'].items():
        paths = [expand_home_path(path) for path in paths.split()]
        paths = list(set(paths))
        paths.sort()
        config['Targets'][label] = '\n' + '\n'.join(paths)



from configparser import ConfigParser
import os
from syndot.init_config import CONFIG_DIR_PATH


COLORSCHEME_PATH = os.path.join(
    CONFIG_DIR_PATH,
    'colorschemes',
    'default.colorscheme'
)
if not os.path.exists(CONFIG_DIR_PATH) or not os.path.exists(COLORSCHEME_PATH):
    COLORSCHEME_PATH = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        '_colorschemes',
        'default.colorscheme'
    )
color_config = ConfigParser()
color_config.read(COLORSCHEME_PATH)
COLOR_MAP = {
    'link': 'color13',
    'symbol': 'color11',
    'settings': 'color08',
    'error': 'color05',
    'prompt_sentence': 'color07',
    'prompt_foreground': 'color07',
    'prompt_background': 'color24',
    'indicator_foreground': 'color12',
    'match_foreground': 'color07',
    'selected_indicator_foreground': 'color09'
}


def _get_color(category: str, fmt: str) -> str:
    hex_color = color_config.get(section='Colors', option=COLOR_MAP[category])
    if fmt == 'HEX':
        return hex_color
    elif fmt == 'RGB':
        r, g, b = tuple(
            int(hex_color.lstrip('#')[i:i+2], 16) for i in (0, 2, 4)
        )
        return f"{r};{g};{b}"
    else:
        raise ValueError("Invalid color format: select 'HEX' or 'RGB'.")


def _rgb_to_ansi(rgb_color: str) -> str:
    return f"\x1b[38;2;{rgb_color}m"


class Color:
    LINK_COLOR = _rgb_to_ansi(_get_color(category='link', fmt='RGB'))
    SYMBOL_COLOR = _rgb_to_ansi(_get_color(category='symbol', fmt='RGB'))
    SETTINGS_COLOR = _rgb_to_ansi(_get_color(category='settings', fmt='RGB'))
    ERROR_COLOR = _rgb_to_ansi(_get_color(category='error', fmt='RGB'))
    PROMPT_SENTENCE = _get_color(category='prompt_sentence', fmt='HEX')
    PROMPT_FOREGROUND = _get_color(category='prompt_foreground', fmt='HEX')
    PROMPT_BACKGROUND = _get_color(category='prompt_background', fmt='HEX')
    INDICATOR_FOREGROUND = _get_color(
        category='indicator_foreground',
        fmt='HEX'
    )
    MATCH_FOREGROUND = _get_color(category='match_foreground', fmt='HEX')
    SELECTED_INDICATOR_FOREGROUND = _get_color(
        category='selected_indicator_foreground',
        fmt='HEX'
    )
    BOLD_START_SEQUENCE = '\033[1m'
    BOLD_END_SEQUENCE = '\033[0m'
    COLOR_END_SEQUENCE = '\x1b[0m'

    @classmethod
    def link(cls, link: str) -> str:
        return cls.LINK_COLOR + link + cls.COLOR_END_SEQUENCE

    @classmethod
    def symbol(cls, symbol: str) -> str:
        return cls.SYMBOL_COLOR + symbol + cls.COLOR_END_SEQUENCE

    @classmethod
    def settings(cls, settings: str) -> str:
        return cls.SETTINGS_COLOR + settings + cls.COLOR_END_SEQUENCE

    @classmethod
    def error(cls, error: str) -> str:
        return cls.ERROR_COLOR + error + cls.COLOR_END_SEQUENCE

    @classmethod
    def highlight(cls, sentence: str) -> str:
        return cls.BOLD_START_SEQUENCE + sentence + cls.BOLD_END_SEQUENCE



__all__ = [
    "add_parser",
    "diffuse_parser",
    "init_parser",
    "link_parser",
    "list_parser",
    "remove_parser",
    "rename_parser",
    "unlink_parser",
]

from argparse import SUPPRESS
from syndot.parser.add_parser import add_parser
from syndot.parser.diffuse_parser import diffuse_parser
from syndot.parser.init_parser import init_parser
from syndot.parser.link_parser import link_parser
from syndot.parser.list_parser import list_parser
from syndot.parser.remove_parser import remove_parser
from syndot.parser.rename_parser import rename_parser
from syndot.parser.unlink_parser import unlink_parser

help_option_parsers = [
    add_parser,
    diffuse_parser,
    init_parser,
    link_parser,
    list_parser,
    remove_parser,
    rename_parser,
    unlink_parser
]

map_option_parsers = [
    add_parser,
    diffuse_parser,
    link_parser,
    list_parser,
    remove_parser,
    rename_parser,
    unlink_parser
]

start_option_parsers = [
    diffuse_parser,
    link_parser,
    unlink_parser
]

confirmation_option_parsers = [
    diffuse_parser,
    link_parser,
    unlink_parser
]

for parser in [
    add_parser,
    diffuse_parser,
    init_parser,
    link_parser,
    list_parser,
    remove_parser,
    rename_parser,
    unlink_parser
]:
    if parser in help_option_parsers:
        parser.add_argument(
            '-h', '--help',
            action='help',
            default=SUPPRESS,
            help="Show this help message and exit"
        )

    if parser in map_option_parsers:
        parser.add_argument(
            '-m', '--mapfile',
            required=False,
            metavar='<MAP_FILE>',
            help="Path to the %(metavar)s. If not provided search for a "
                 "'map.ini' file in the current directory, so not required if "
                 "the current directory is the settings directory"
        )

    if parser in start_option_parsers:
        parser.add_argument(
            '-s', '--start',
            required=False,
            dest='start',
            metavar='<PATH_START>',
            help="Filter target based on path starting with <PATH_START>"
        )

    if parser in confirmation_option_parsers:
        parser.add_argument(
            '-n', '--no-confirm',
            action='store_true',
            default=False,
            required=False,
            dest='no_confirm',
            help="Do not ask for confirmation"
        )



from argparse import Namespace
import os
from syndot.init_config import LOG_FILE_PATH
from syndot.utils.commands import skip_dotfiles, print_dotfiles_to_manage
from syndot.utils.file_actions import remove
from syndot.utils.logger import log_error
from syndot.utils.map_file import get_map_info, read_map_file
from syndot.utils.path import compose_target_paths
from syndot.utils.print_ import (
    print_action,
    print_error,
    print_highlight
)
from syndot.utils.prompt import ask_to_proceed


def diffuse(args: Namespace) -> None:
    settings_dir, targets = get_map_info(
        config=read_map_file(map_file_path=args.mapfile),
        args=args
    )

    targets_to_be_diffused = {}
    already_existing_system = {}
    already_diffused_targets = []
    wrong_existing_links = {}
    missing_settings_targets = []
    settings_are_links = []

    print_highlight("Looking for files and directories to diffuse...")

    for target in targets:
        system_target_path, settings_target_path = compose_target_paths(
            settings_dir=settings_dir,
            target=target
        )

        if _check_targets_to_be_diffused(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            targets_to_be_diffused[system_target_path] = settings_target_path

        if _check_already_existing_system(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            already_existing_system[system_target_path] = settings_target_path

        if _check_already_diffused_targets(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            already_diffused_targets.append(system_target_path)

        if _check_wrong_existing_links(
            system_target_path=system_target_path,
            settings_target_path=settings_target_path
        ):
            wrong_existing_links[system_target_path] = settings_target_path

        if _check_missing_settings_targets(
            settings_target_path=settings_target_path
        ):
            missing_settings_targets.append(settings_target_path)

        if _check_settings_are_links(
            settings_target_path=settings_target_path
        ):
            settings_are_links.append(settings_target_path)

    if not any([
        targets_to_be_diffused,
        already_existing_system,
        already_diffused_targets,
        wrong_existing_links,
        missing_settings_targets,
        settings_are_links
    ]):
        print_highlight("No files or directories found to diffuse.")

    diffuse_dotfiles(
        targets_list=targets_to_be_diffused,
        many_targets_sentence=f"Found {len(targets_to_be_diffused.keys())} "
                              f"files or directories to diffuse.\nSymbolic "
                              f"links to these settings files and directories "
                              f"will be created in the respective system "
                              f"directories.",
        single_file_sentence=f"Found {len(targets_to_be_diffused.keys())} "
                             f"file to diffuse.\nA symbolic link to this file "
                             f"will be created in the respective system "
                             f"directory.",
        single_directory_sentence=f"Found {len(targets_to_be_diffused.keys())}"
                                  f" directory to diffuse.\nA symbolic link "
                                  f"to this directory will be created in the "
                                  f"respective system directory.",
        remove_system=False,
        ask_for_confirmation=not args.no_confirm
    )

    diffuse_dotfiles(
        targets_list=already_existing_system,
        many_targets_sentence=f"Found {len(already_existing_system.keys())} "
                              f"already existing system files or "
                              f"directories.\nThey will be removed and "
                              f"replaced by symbolic links to the respective "
                              f"files and directories in the settings "
                              "directory.",
        single_file_sentence=f"Found {len(already_existing_system.keys())} "
                             f"already existing system file.\nIt will be "
                             f"removed and replaced by a symbolic link to the "
                             f"respective file in the settings directory.",
        single_directory_sentence=f"Found "
                                  f"{len(already_existing_system.keys())} "
                                  f"already existing system directory.\nIt "
                                  f"will be removed and replaced by a "
                                  f"symbolic link to the respective directory "
                                  f"in the settings directory.",
        remove_system=True,
        ask_for_confirmation=not args.no_confirm
    )

    diffuse_dotfiles(
        targets_list=wrong_existing_links,
        many_targets_sentence=f"Found {len(wrong_existing_links.keys())} "
                              f"files or directories that are links to wrong "
                              f"files or directories.\nThey will be removed "
                              f"and replaced by symbolic links to the "
                              f"respective files and directories in the "
                              f"settings directory.",
        single_file_sentence=f"Found {len(wrong_existing_links.keys())} file "
                             f"that is a link to a wrong file.\nIt will be "
                             f"removed and replaced by a symbolic link to the "
                             f"respective file in the settings directory.",
        single_directory_sentence=f"Found {len(wrong_existing_links.keys())} "
                                  f"directory that is a link to a wrong "
                                  f"directory.\nIt will be removed and "
                                  f"replaced by a symbolic link to the "
                                  f"respective directory in the settings "
                                  f"directory.",
        remove_system=True,
        ask_for_confirmation=not args.no_confirm
    )

    skip_dotfiles(
        targets_list=already_diffused_targets,
        many_targets_sentence=f"Skipping {len(already_diffused_targets)} "
                              f"already diffused files or directories:",
        single_file_sentence=f"Skipping {len(already_diffused_targets)} "
                             f"already diffused file:",
        single_directory_sentence=f"Skipping {len(already_diffused_targets)} "
                                  f"already diffused directory:"
    )

    skip_dotfiles(
        targets_list=missing_settings_targets,
        many_targets_sentence=f"Skipping {len(missing_settings_targets)} "
                              f"missing settings files or directories:",
        single_file_sentence=f"Skipping {len(missing_settings_targets)} "
                             f"missing settings file:",
        single_directory_sentence=f"Skipping {len(missing_settings_targets)} "
                                  f"missing settings directory:"
    )

    skip_dotfiles(
        targets_list=settings_are_links,
        many_targets_sentence=f"Skipping {len(settings_are_links)} settings "
                              f"files or directories because are links:",
        single_file_sentence=f"Skipping {len(settings_are_links)} settings "
                             f"file because is a link:",
        single_directory_sentence=f"Skipping {len(settings_are_links)} "
                                  f"settings directory because is a link:"
    )


def diffuse_dotfiles(
    targets_list: dict[str, str],
    many_targets_sentence: str,
    single_file_sentence: str,
    single_directory_sentence: str,
    remove_system: bool,
    ask_for_confirmation: bool
) -> None:
    if targets_list:
        n_targets = len(targets_list.keys())

        print_dotfiles_to_manage(
            targets_list=targets_list,
            many_targets_sentence=many_targets_sentence,
            single_file_sentence=single_directory_sentence,
            single_directory_sentence=single_directory_sentence,
            symbol='<--'
        )

        proceed = ask_to_proceed() if ask_for_confirmation else True

        if proceed:
            for i, (system_target_path, settings_target_path) in \
                    enumerate(targets_list.items(), 1):
                try:
                    print_action(
                        action_type='diffuse',
                        system_target_path=system_target_path,
                        settings_target_path=settings_target_path
                    )
                    print(f"Total ({i}/{n_targets})", end='\r')

                    if remove_system:
                        remove(system_target_path)
                    if not os.path.exists(os.path.dirname(system_target_path)):
                        os.makedirs(os.path.dirname(system_target_path))
                    os.symlink(settings_target_path, system_target_path)
                except Exception as e:
                    print_error(
                        f"Error in diffusing {system_target_path}. "
                        f"Check {LOG_FILE_PATH} for more details."
                    )
                    log_error(f"{e}")
        print('\n')


def _check_targets_to_be_diffused(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return not os.path.islink(settings_target_path) and \
        os.path.exists(settings_target_path) and \
        not os.path.islink(system_target_path) and \
        not os.path.exists(system_target_path)


def _check_already_existing_system(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return not os.path.islink(settings_target_path) and \
        os.path.exists(settings_target_path) and \
        not os.path.islink(system_target_path) and \
        os.path.exists(system_target_path)


def _check_already_diffused_targets(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return not os.path.islink(settings_target_path) and \
        os.path.exists(settings_target_path) and \
        os.path.islink(system_target_path) and \
        (os.readlink(system_target_path) == settings_target_path)


def _check_wrong_existing_links(
    system_target_path: str,
    settings_target_path: str
) -> bool:
    return not os.path.islink(settings_target_path) and \
        os.path.exists(settings_target_path) and \
        os.path.islink(system_target_path) and \
        (os.readlink(system_target_path) != settings_target_path)


def _check_missing_settings_targets(settings_target_path: str) -> bool:
    return not os.path.islink(settings_target_path) and \
        not os.path.exists(settings_target_path)


def _check_settings_are_links(settings_target_path: str) -> bool:
    return os.path.islink(settings_target_path)



from setuptools import setup
import subprocess


version = subprocess.run(
    'git describe --tags'.split(),
    stdout=subprocess.PIPE
).stdout.decode('utf-8').strip()

try:
    setup(version=version)
except Exception as e:
    if e.args[0].startswith("Invalid version: "):
        setup(version='v0.0.0')
    else:
        print(e)



import os
import shutil
from syndot.utils.path import expand_home_path


def copy(source: str, destination: str) -> None:
    destination_parent = os.path.dirname(destination)
    if not os.path.exists(destination_parent):
        os.makedirs(destination_parent)

    if os.path.isfile(source):
        shutil.copy2(source, destination)
    elif os.path.isdir(source):
        shutil.copytree(source, destination, symlinks=True)
        st = os.stat(source)
        os.chown(destination, st.st_uid, st.st_gid)


def change_parent_owner(
    source: str,
    destination: str,
    settings_dir: str
) -> None:
    protected_directories = (
        expand_home_path(settings_dir),
        expand_home_path('~'),
        '/'
    )
    while destination not in protected_directories:
        st = os.stat(source)
        try:
            os.chown(destination, st.st_uid, st.st_gid)
        except PermissionError:
            pass
        source = os.path.dirname(source)
        destination = os.path.dirname(destination)


def change_child_owner(source: str, destination: str) -> None:
    source_content = os.listdir(source)
    destination_content = os.listdir(destination)
    for source_target, destination_target in \
            zip(source_content, destination_content):
        source_target_path = os.path.join(source, source_target)
        destination_target_path = os.path.join(destination, destination_target)
        st = os.stat(source_target_path)
        os.chown(destination_target_path, st.st_uid, st.st_gid)
        if os.path.isdir(destination_target_path):
            change_child_owner(
                source=source_target_path,
                destination=destination_target_path
            )


def remove(path: str) -> None:
    if os.path.islink(path):
        os.unlink(path)
    else:
        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)



from argparse import Namespace
from syndot.utils.map_file import read_map_file, write_map_file
from syndot.utils.path import expand_home_path


def rename(args: Namespace) -> None:
    if args.old_label == args.new_label:
        raise NameError("The new label must be different from the old one")

    map_file_path = expand_home_path(
        args.mapfile if args.mapfile is not None else 'map.ini'
    )

    config = read_map_file(map_file_path=map_file_path)

    targets = dict(config['Targets'])

    if args.old_label in targets.keys():
        if args.new_label in targets.keys():
            paths = [*targets[args.old_label].split(),
                     *targets[args.new_label].split()]
            paths.sort()
            targets[args.new_label] = '\n' + '\n'.join(paths)
            targets.pop(args.old_label)
        else:
            targets[args.new_label] = targets.pop(args.old_label)
    else:
        raise KeyError(f"Label {args.old_label!r} not found in the map file")

    config['Targets'] = dict(sorted(targets.items()))

    write_map_file(map_file_path=map_file_path, config=config)



from argparse import ArgumentParser, HelpFormatter, SUPPRESS, _SubParsersAction
from operator import attrgetter
from syndot.version import __version__


class GeneralFormatter(HelpFormatter):

    def __init__(self, prog):
        super().__init__(prog, max_help_position=20, width=80)

    def _iter_indented_subactions(self, action):
        if hasattr(action, '_get_subactions'):
            get_subactions = action._get_subactions
            if isinstance(action, _SubParsersAction):
                for subaction in sorted(
                    get_subactions(),
                    key=lambda x: x.dest
                ):
                    yield subaction
            else:
                for subaction in get_subactions():
                    yield subaction

    def _format_action(self, action):
        result = super()._format_action(action)
        if isinstance(action, _SubParsersAction):
            return "%*s%s" % (self._current_indent, '', result.lstrip())
        return result

    def _format_action_invocation(self, action):
        if isinstance(action, _SubParsersAction):
            return ""
        return super()._format_action_invocation(action)


class CommandFormatter(HelpFormatter):

    def __init__(self, prog):
        super().__init__(prog, max_help_position=40, width=80)

    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter('option_strings'))
        super(CommandFormatter, self).add_arguments(actions)

    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ', '.join(action.option_strings) + ' ' + args_string


parser = ArgumentParser(
    prog='syndot',
    usage="%(prog)s <COMMAND> [<OPTIONS>...]",
    description="Manage symlinks to dotfiles",
    epilog="Config file path: ~/.config/syndot",
    add_help=False,
    formatter_class=GeneralFormatter
)

parser.add_argument(
    '-h', '--help',
    action='help',
    default=SUPPRESS,
    help="Show this help message and exit"
)

parser.add_argument(
    '-v', '--version',
    action='version',
    version=f"%(prog)s {__version__}",
    help="Show program\"s version number and exit"
)

command_parser = parser.add_subparsers(
    dest='command',
    title='commands'
)



import logging
from syndot.init_config import LOG_FILE_PATH


def log_error(error_message: str) -> None:
    logging.basicConfig(
        filename=LOG_FILE_PATH,
        level=logging.ERROR,
        format="%(asctime)s.%(msecs)03d %(levelname)s %(message)s",
        datefmt="%Y-%m-%d,%H:%M:%S"
    )
    logger = logging.getLogger(__name__)
    logger.error(error_message)
